library(s2dv)
library(multiApply)
library(CSTools)
library(CSIndicators)
setwd("/esarchive/scratch/eduzenli/gitlab/sunset")
source("modules/Loading/Loading.R")
source("modules/Downscaling/Downscaling.R")
source("modules/Skill/Skill.R")
source("modules/Anomalies/Anomalies.R")
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/Interpolation.R')
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/Intbc.R')
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/Intlr.R')
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/Analogs.R')
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/LogisticReg.R')
source('/esarchive/scratch/eduzenli/gitlab/csdownscale/R/Utils.R')

REF <- "CERRA" ## reference for the local scale var. for lsvar the reference is ERA5.
savedata <- TRUE
vars <- c("tas")  # 'tas', 'tasmax', 'tasmin', 'prlr','sfcWind'
freq <- "daily_mean" ## daily, daily_mean or weekly_mean
dwn_type <- 'analogs'   ### 'int', 'intbc', 'intlr', 'analogs', 'logreg'
int_method <- NULL  # c('bic','con') # "NULL",'con', 'bil', 'bic', 'nn', 'con2',"dis","laf"
bc_method <- NULL # # NULL, "quantile_mapping", "bias", "evmos", "mse_min", "crps_min", "rpc-based", "qm" 
lr_method <- NULL             # 'NULL' 'basic', 'large-scale', '4nn'
log_reg_method <- NULL #  "NULL" ,"ens_mean", "ens_mean_sd", "sorted_members"
nanalogs <- 1
approach <- c("MOS", "PP", "WR", "WR_EOF")  ##  MOS, PP, WR, WR_EOF
ncenters <- 4  ## howmany cluster for WR or WR_EOF. you cannot select multiple.  For MOS no needed.
lsvariable <- "psl" ## which large scale variable will be used for PP, WR, WR_EOF. For MOS no needed.
latmin = 30 # large-scale variable. For MOS no needed.
latmax = 65 # large-scale variable. For MOS no needed.
lonmin = -30 # large-scale variable. For MOS no needed.
lonmax = 25 # large-scale variable. For MOS no needed.

outpath <- paste0("/esarchive/scratch/eduzenli/downscaling/aspect/spring_frost/", dwn_type, 
                  "/data/cerra/")

skillpath <- paste0("/esarchive/scratch/eduzenli/downscaling/aspect/spring_frost/", dwn_type, 
                    "/skill/cerra/")


## Specify different downscaling scenarios
METHODS_DUM<-list(int_method,bc_method,lr_method,log_reg_method,nanalogs)
names(METHODS_DUM)<-c("int_method","bc_method","lr_method","log_reg_method","nanalogs")
METHODS<-METHODS_DUM[!as.logical(lapply(METHODS_DUM, is.null))]
method_loop<-expand.grid(METHODS)

naming_1<-names(method_loop)[1]

if(length(names(method_loop))==2)
{naming_2<-names(method_loop)[2]} else
{naming_2<-NULL}


args <- commandArgs(trailingOnly = TRUE)
target_month <- as.numeric(args[1])
lt <- as.numeric(args[2])
diffmon <- target_month - lt
smonth <- paste0(sprintf('%02d', ifelse(diffmon > 0, diffmon, 12 + diffmon)), '01')

recipe_file <- paste0("/esarchive/scratch/eduzenli/gitlab/aspect/",
                        "recipe_spring_frost_downscaling_seasonal.yml")

recipe <- prepare_outputs(recipe_file)
recipe$Analysis$Time$sdate <- smonth
size <- recipe$Analysis$Workflow$Downscaling$size
if (!is.null(size) & dwn_type == 'analogs') {
  smonth_dum<-smonth
  smonth<-numeric(3)
  for (i in 1:3) {
    smonth[i] <- as.numeric(substr(smonth_dum, start = 1, stop = 2))+ (i-3+1)
  }
  smonth[smonth<1] <- smonth[smonth<1]+12
  smonth[smonth>12] <- smonth[smonth>12]-12
  smonth <- paste0(sprintf('%02d', smonth), '01')
}

recipe$Analysis$Time$ftime_min <- lt+1
recipe$Analysis$Time$ftime_max <- lt+1
recipe$Analysis$Variables$freq <- freq


if ( any(toupper(approach) != "MOS") ) {

  recipe$Analysis$Region$latmin <- latmin
  recipe$Analysis$Region$latmax <- latmax
  recipe$Analysis$Region$lonmin <- lonmin
  recipe$Analysis$Region$lonmax <- lonmax
  recipe$Analysis$Variables$name <- lsvariable
  recipe$Analysis$Datasets$Reference$name <- "ERA5"

  # analog extended window
  if (!is.null(size) & dwn_type == 'analogs') {

    for (dd in 1:length(smonth)) {
      recipe$Analysis$Time$sdate <-smonth[dd]
      assign(paste0("data_",dd),load_datasets(recipe))

      k<-get(paste0("data_",dd))

      ## fix all months obs to 31 days
      if (dim(k$obs$data)["time"]!=31) {

        dim<-dim(k$obs$data)
        dim["time"]<- 31-dim["time"]
        dum<-array(NA,dim)
        old<-k$obs$data
        k$obs$data<-abind(old,dum,along=6)
        names(dim(k$obs$data))<-names(dim(dum))

        assign(paste0("data_",dd),k)
      }
    }

    dataL <- data_2
    dataL$obs$data <- abind(data_1$obs$data,data_2$obs$data,data_3$obs$data,along=which(names(dim(data_1$obs$data))=="time"))

    names(dim(dataL$obs$data)) <- names(dim(data_2$obs$data))

    dim(dataL$obs$data)<-c(dim(dataL$obs$data)[c("dat","var","sday","sweek","syear")],time=31,smonth=3,dim(data$obs$data)[c("latitude","longitude","ensemble")])

    rm (data_1,data_2,data_3)
  } else {

    recipe$Analysis$Time$sdate <-smonth
    dataL <- load_datasets(recipe)

  }

  # compute anomalies
  clim_hcst <- Apply(dataL$hcst$data, target_dims = c('syear','ensemble'), mean, na.rm = T)$output1
  clim_obs <- Apply(dataL$obs$data, target_dims = c('syear','ensemble'), mean, na.rm = T)$output1

  dataL$hcst$data <- Ano(data = dataL$hcst$data, clim = clim_hcst)
  dataL$obs$data <- Ano(data = dataL$obs$data, clim = clim_obs)

  syear_num <- dim(dataL$obs$data) [names(dim(dataL$obs$data)) == "syear"]
  day_num <- dim(dataL$obs$data) [names(dim(dataL$obs$data)) == "time"]
  ens_num <- dim(dataL$hcst$data) [names(dim(dataL$hcst$data)) == "ensemble"]

}


if ( any(toupper(approach) == "WR") | any(toupper(approach) == "WR_EOF") ) {

  ## OBS WRs
 
  if (suppressWarnings(!.check_coords(lat1 = as.numeric(dataL$obs$coords$lat),
                                      lat2 = as.numeric(dataL$hcst$coords$lat),
                                      lon1 = as.numeric(dataL$obs$coords$lon),
                                      lon2 = as.numeric(dataL$hcst$coords$lon)))) {

    obsL_interpolated <- suppressWarnings(Interpolation(exp = dataL$obs$data,
                                                        lats = dataL$obs$coords$lat,
                                                        lons = dataL$obs$coords$lon,
                                                        target_grid = dataL$hcst$attrs$source_files[1],
                                                        lat_dim = "latitude",
                                                        lon_dim = "longitude", region = NULL,
                                                        method_remap = "conservative", 
                                                        ncores = recipe$Analysis$ncores))

    obsL_interpolated <- suppressWarnings(s2dv_cube(data = obsL_interpolated$data,
                                                    coords = list(latitude = obsL_interpolated$lat,
                                                    longitude = obsL_interpolated$lon)))
  } else {
    obsL_interpolated <- dataL$obs
  }

  names(dim(obsL_interpolated$data)) [which(names(dim(obsL_interpolated$data))=="latitude")] <- "lat"
  names(dim(obsL_interpolated$data)) [which(names(dim(obsL_interpolated$data))=="longitude")] <- "lon"

  obsL_interpolated$dims <- dim(obsL_interpolated$data)
  obsL_interpolated$coords$lat <- obsL_interpolated$coords$latitude
  obsL_interpolated$coords$lon <- obsL_interpolated$coords$longitude
  obsL_interpolated$coords$latitude <- NULL
  obsL_interpolated$coords$longitude <- NULL
  obsL_interpolated$data <- MergeDims(obsL_interpolated$data, 
                                      merge_dims=c("time","syear"), rename_dim = "time")
  ### WR
  WR_obsL <- CST_WeatherRegimes(data = obsL_interpolated, EOFs = FALSE, 
                                ncenters = ncenters, ncores = recipe$Analysis$ncores)

  ### WR_EOF
  time_length <- dim(obsL_interpolated$data)[names(dim(obsL_interpolated$data)) == "time"]

  dum <- EOF(obsL_interpolated$data, lat = obsL_interpolated$coords$lat,
             lon = obsL_interpolated$coords$lon,
             neof = time_length,
             time_dim = "time",
             space_dim = c("lat", "lon"))
  neof <- min(which(cumsum(dum$var) > 95))

  WR_obsL_eof <- CST_WeatherRegimes(data = obsL_interpolated, EOFs = TRUE,
                                      ncenters = ncenters, neof = neof, 
                                      ncores = recipe$Analysis$ncores)

  ## EXP WRs

  if (suppressWarnings(!.check_coords(lat1 = as.numeric(obsL_interpolated$coords$lat),
                                      lat2 = as.numeric(dataL$hcst$coords$lat),
                                      lon1 = as.numeric(obsL_interpolated$coords$lon),
                                      lon2 = as.numeric(dataL$hcst$coords$lon)))) {
    hcstL_interpolated <- suppressWarnings(Interpolation(exp = dataL$hcst$data,
                                                         lats = as.numeric(dataL$hcst$coords$lat),
                                                         lons = as.numeric(dataL$hcst$coords$lon),
                                                         target_grid = dataL$hcst$attrs$source_files[1],
                                                         lat_dim = "latitude",
                                                         lon_dim = "longitude",
                                                         method_remap = "conservative",
                                                         region = NULL, 
                                                         ncores = recipe$Analysis$ncores))

    hcstL_interpolated <- suppressWarnings(s2dv_cube(data = hcstL_interpolated$data,
                                                     coords = list(latitude = hcstL_interpolated$lat,
                                                     longitude = hcstL_interpolated$lon)))
  } else {
    hcstL_interpolated <- data$hcst
  }

  names(dim(hcstL_interpolated$data)) [which(names(dim(hcstL_interpolated$data))=="latitude")] <- "lat"
  names(dim(hcstL_interpolated$data)) [which(names(dim(hcstL_interpolated$data))=="longitude")] <- "lon"
  hcstL_interpolated$dims <- dim(hcstL_interpolated$data)
  hcstL_interpolated$coords$lat <- hcstL_interpolated$coords$latitude
  hcstL_interpolated$coords$lon <- hcstL_interpolated$coords$longitude
  hcstL_interpolated$coords$latitude <- NULL
  hcstL_interpolated$coords$longitude <- NULL
  hcstL_interpolated$data <- MergeDims(hcstL_interpolated$data,
                                       merge_dims=c("time","syear"), rename_dim = "time")

  ### WR
  WR_hcstL <- CST_RegimesAssign(data = hcstL_interpolated,
                                ref_maps = WR_obsL,
                                method = 'distance', composite = TRUE,
                                memb = FALSE)

  #### dimension ordering hcstL
  dum <- SplitDim(WR_hcstL$statistics$cluster, split_dim = "time", 
                  indices = rep(1:day_num, syear_num), new_dim_name = c("syear")  )

  time_index <- which(sapply(names(dim(dum)), function(dim_name) "time" %in% dim_name))
  syear_index <- which(sapply(names(dim(dum)), function(dim_name) "syear" %in% dim_name))
  names(dim(dum))[[time_index]] <- "syear"
  names(dim(dum))[[syear_index]] <- "time"

  WR_hcstL$statistics$cluster <-dum

  #### dimension ordering obsL
  dum <- SplitDim(WR_obsL$statistics$cluster, split_dim = "time",
                  indices = rep(1:day_num, syear_num), new_dim_name = c("syear"))

  time_index <- which(sapply(names(dim(dum)), function(dim_name) "time" %in% dim_name))
  syear_index <- which(sapply(names(dim(dum)), function(dim_name) "syear" %in% dim_name))
  names(dim(dum))[[time_index]] <- "syear"
  names(dim(dum))[[syear_index]] <- "time"

  WR_obsL$statistics$cluster <-dum

  ### WR_EOF
  WR_hcstL_eof <- CST_RegimesAssign(data = hcstL_interpolated,
                                  ref_maps = WR_obsL_eof,
                                  method = 'distance', composite = TRUE,
                                  memb = FALSE)

  #### dimension ordering hcstL_eof
  dum <- SplitDim(WR_hcstL_eof$statistics$cluster, split_dim = "time", 
                  indices = rep(1:day_num, syear_num), new_dim_name = c("syear")  )

  time_index <- which(sapply(names(dim(dum)), function(dim_name) "time" %in% dim_name))
  syear_index <- which(sapply(names(dim(dum)), function(dim_name) "syear" %in% dim_name))
  names(dim(dum))[[time_index]] <- "syear"
  names(dim(dum))[[syear_index]] <- "time"

  WR_hcstL_eof$statistics$cluster <-dum

  #### dimension ordering obsL_eof
  dum <- SplitDim(WR_obsL_eof$statistics$cluster, split_dim = "time",
                  indices = rep(1:day_num, syear_num), new_dim_name = c("syear"))

  time_index <- which(sapply(names(dim(dum)), function(dim_name) "time" %in% dim_name))
  syear_index <- which(sapply(names(dim(dum)), function(dim_name) "syear" %in% dim_name))
  names(dim(dum))[[time_index]] <- "syear"
  names(dim(dum))[[syear_index]] <- "time"

  WR_obsL_eof$statistics$cluster <-dum


}


# LOCAL SCALE VARIABLE AND DOWNSCALING

#recipe$Analysis$Region$latmin <- 39
#recipe$Analysis$Region$latmax <- 44
#recipe$Analysis$Region$lonmin <- -1
#recipe$Analysis$Region$lonmax <- 4

recipe$Analysis$Region$latmin <- 40.5
recipe$Analysis$Region$latmax <- 42.5
recipe$Analysis$Region$lonmin <- .5
recipe$Analysis$Region$lonmax <- 2.5

for (var in vars) {

  recipe$Analysis$Variables$name <- var
  recipe$Analysis$Datasets$Reference$name <- REF

  if (var=="prlr" ) {
    recipe$Analysis$Datasets$Reference$name <- "CERRA-Land"
  }

  if (recipe$Analysis$Variables$freq == "daily_mean" &  (var == "tasmin" | var == "tasmax")) {
    recipe$Analysis$Variables$freq <- "daily"
  }

  # Load datasets

  # analog extended window

  if (!is.null(size) & dwn_type == 'analogs') {

  for (dd in 1:length(smonth)) {

    recipe$Analysis$Time$sdate <-smonth[dd]
    assign(paste0("data_",dd),load_datasets(recipe))

    k<-get(paste0("data_",dd))

    ## fix all months obs to 31 days
    if (dim(k$obs$data)["time"]!=31) {

      dim<-dim(k$obs$data)
      dim["time"]<- 31-dim["time"]
      dum<-array(NA,dim)
      old<-k$obs$data
      k$obs$data<-abind(old,dum,along=6)
      names(dim(k$obs$data))<-names(dim(dum))

      assign(paste0("data_",dd),k)
    }
  }

  data <- data_2

  data$obs$data <- abind(data_1$obs$data,data_2$obs$data,data_3$obs$data,along=which(names(dim(data_1$obs$data))=="time"))

  names(dim(data$obs$data)) <- names(dim(data_2$obs$data))

  dim(data$obs$data)<-c(dim(data$obs$data)[c("dat","var","sday","sweek","syear")],time=31,smonth=3,dim(data$obs$data)[c("latitude","longitude","ensemble")])

  rm (data_1,data_2,data_3)
  } else {

  recipe$Analysis$Time$sdate <-smonth
  data <- load_datasets(recipe)

  }

  # compute anomalies
  clim_hcst <- Apply(data$hcst$data, target_dims = c('syear','ensemble'), mean, na.rm = T)$output1
  clim_obs <- Apply(data$obs$data, target_dims = c('syear','ensemble'), mean, na.rm = T)$output1

  data$hcst$data <- Ano(data = data$hcst$data, clim = clim_hcst)
  data$obs$data <- Ano(data = data$obs$data, clim = clim_obs)

  for (mm in 1:nrow(method_loop)) {

    # prepare the (mm)th scenario to arrange the recipe
    assign(names(method_loop)[1],as.character(method_loop[mm,1]))
    assign(names(method_loop)[2],as.character(method_loop[mm,2]))

    #prepare the file name
    ifelse(is.null(naming_2), NN <- paste0(get(naming_1),"-"),
           NN <- paste0(get(naming_1),"-",get(naming_2),"-"))

    # Modify the recipe manually
    recipe$Analysis$Workflow$Downscaling$type           <- dwn_type
    recipe$Analysis$Workflow$Downscaling$int_method     <- int_method
    recipe$Analysis$Workflow$Downscaling$bc_method      <- bc_method
    recipe$Analysis$Workflow$Downscaling$lr_method      <- lr_method
    recipe$Analysis$Workflow$Downscaling$log_reg_method <- log_reg_method
    recipe$Analysis$Workflow$Downscaling$nanalogs       <- nanalogs

    for (app in approach) {

      data$obsL <- NULL
      data$hcstL <- NULL
      cnum <- "" # for file naming
      if ( toupper(app) == "PP" ) {
        if ( dwn_type != "analogs" ) {  ## PP is only valid for analogs AT THIS POINT
          stop("PP approach is only valid for the analogs method.'")
        }
        data$obsL <- dataL$obs
        data$hcstL <- dataL$hcst
      }

      # Downscale datasets

      if ( !(toupper(app) == "WR" | toupper(app) == "WR_EOF") ) {
        downscaled_data <- Downscaling(recipe, data)
      } else {     #### WEATHER REGIME DOWNSCALING
        cnum <- ncenters # for file naming
        if ( toupper(app) == "WR" ) {
          class_hcstL <- WR_hcstL$statistics$cluster
          class_obsL <- WR_obsL$statistics$cluster
        } else if (toupper(app) == "WR_EOF") {
          class_hcstL <- WR_hcstL_eof$statistics$cluster
          class_obsL <- WR_obsL_eof$statistics$cluster
        }

        for (cc in 1:ncenters) {

          dum <- data

          ## assign NA if    
          dum$hcst$data <- Apply(list(dum$hcst$data, class_hcstL),
                                 target_dims = c( "syear", "ensemble", "time"),
                                 fun = function (x, y) {
                                 d <- which(y != cc)
                                 x [d]<- NA
                                 return(x)}
                                )$output1

          dum$obs$data <- Apply(list(dum$obs$data, class_obsL), target_dims = c( "syear", "time"),
                                fun = function (x, y) {
                                d <- which(y != cc)
                                x [d]<- NA
                                return(x)}
                               )$output1

          assign(paste0("data_", cc), dum)

          dum <- Downscaling(recipe, get(paste0("data_", cc)))

          assign (paste0("downscaled_data_", cc), dum)

          print (paste0(cc,"/",ncenters," is completed"))
        } 

        files <- ls(pattern = "downscaled_data_") ## downscaled files for different classes
        downscaled_data <- downscaled_data_1

        for (nc in 2:ncenters) {
          file <- get(files[[nc]])$hcst$data
          file_obs <- get(files[[nc]])$obs$data
          downscaled_data$hcst$data [is.na(downscaled_data$hcst$data) & !is.na(file)] <-
            file [is.na(downscaled_data$hcst$data) & !is.na(file)]
          downscaled_data$obs$data [is.na(downscaled_data$obs$data) & !is.na(file_obs)] <-
            file_obs [is.na(downscaled_data$obs$data) & !is.na(file_obs)]
        }
      }

      if(recipe$Analysis$Horizon == "seasonal" &
        (recipe$Analysis$Variables$freq == "daily_mean" |
         recipe$Analysis$Variables$freq == "daily")) {

         saveRDS(downscaled_data, file = paste0(outpath, var, '-',
                 dwn_type, '-',
                 NN,
                'tm', target_month, '-lt', lt,'_', toupper(app),cnum,'_daily.RDS'))

        downscaled_data$obs <- CST_PeriodMean(downscaled_data$obs, time_dim = "time")
        downscaled_data$hcst <- CST_PeriodMean(downscaled_data$hcst, time_dim = "time")
      }
 
      if(dwn_type == "logreg") {
        prob_obs <- Apply(downscaled_data$obs$data,
                          target_dims = "syear",
                          function(x) {
                          terc <- convert2prob(as.vector(x),
                          prob = c(1/3, 2/3)) },
                          output_dims = c("syear","category"))$output1

        downscaled_data$obs$data<- Subset(prob_obs,
                                          "ensemble",
                                          1,drop="selected")


        rpss <- RPSS(downscaled_data$hcst$data, downscaled_data$obs$data,
                     time_dim = "syear",
                     memb_dim=NULL,
                     cat_dim="category",
                     Fair = FALSE,
                     ncores = recipe$Analysis$ncores)

  #bss10<-RPSS_from_probs(probs_exp,probs_obs,sdate_dim="syear",cat_dim="category",prob_thresholds=.1)
  #bss90<-RPSS_from_probs(probs_exp,probs_obs,sdate_dim="syear",cat_dim="category",prob_thresholds=.9)
        rpss$rpss<-Subset(rpss$rpss,list("time","latitude","longitude"),
                          list(1,1:dim(rpss$rpss)[names(dim(rpss$rpss))=="latitude"],
                          1:dim(rpss$rpss)[names(dim(rpss$rpss))=="longitude"]),
                          drop="non-selected")
        skill_values<-list()
        skill_values<-rpss

      } else {
        skill_values <- Skill(recipe, downscaled_data)
      }

      if (savedata) {

      # Save downscaled data to later intercompare them
      saveRDS(downscaled_data, file = paste0(outpath, var, '-',
              dwn_type, '-',
              NN,
              'tm', target_month, '-lt', lt,'_', toupper(app),cnum,'.RDS'))

      saveRDS(skill_values, file = paste0(skillpath, var, '-',
              dwn_type, '-',
              NN,
             'tm', target_month, '-lt', lt,'_', toupper(app),cnum,'.RDS'))
      }
      print( paste0(var," variable"," is downscaled through the ",
             dwn_type, "-", get(names(method_loop)[1]),"-",
             get(names(method_loop)[2])," method and ",
             app, " approach for lead time ",
             lt , ". The output is saved to the outpath." ))

    }
  }
}


